package org.clazzes.remoting;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.rmi.server.UID;
import java.util.Map;

import org.clazzes.remoting.loading.ClassBytes;

/***********************************************************
 * $Id: RemoteInvoker.java 48 2008-10-29 15:34:24Z wglas $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 14.10.2008
 * 
 ***********************************************************/

/**
 * An invoker for remote methods.
 * 
 * @author wglas
 */
public interface RemoteInvoker extends InvocationContext {

    /**
     * @param subSystem The subsystem, on which the method is invoked.
     * @param arg The argument passed to the invocation handler.
     * @param metadata The metadata accompanying the remote call, which is passed to the
     *                 invocation handler. Metadata may be used to communicate authorization
     *                 information such as user/password or a session Id.
     * @return The result of the operation.
     * 
     * @throws InvocationTargetException If {@link InvocationHandler#invoke(InvocationContext, Object, Map)}
     *                 on the server threw an exception. The exception of the server may
     *                 be retrieved by {@link InvocationTargetException#getCause()}.
     * @throws IOException Upon errors in the transport layer-
     */
    public Object invoke (String subSystem, Object arg, Map<String,Object> metadata)
    throws InvocationTargetException, IOException;

    /**
     * Request a binary class from the server.
     * 
     * @param className The fully qualified name of the class to load.
     * @return The binary class.
     * @throws IOException Upon errors in the transport layer-
     */
    public ClassBytes loadClass(String className) throws IOException;
    
    /**
     * @param handler An invocation handler for the given callback.
     * @return The unique ID of the registered handler. The ID is unique in the
     *         context of the virtual machine of the client, so the server must be
     *         aware of the fact, that multiple clients may register callbacks with
     *         the same client side UID. 
     *         
     * @throws IOException
     */
    public UID registerCallback(InvocationHandler handler) throws IOException;
    
    /**
     * Deregister a callback previously added with {@link #registerCallback(InvocationHandler)}.
     * 
     * @param handlerUID The handler UID of the callback to deregister
     *             returned by {{@link #registerCallback(InvocationHandler)}.
     */
    public void deregisterCallback(UID handlerUID);
    
    /**
     * @return Whether this invoker is connected to the remote resource. If the return
     *         value is <code>true</code>, the server-generated client UID may be
     *         retrieved by {@link #getClientUID()}.
     */
    public boolean isConnected();

    /**
     * Disconnect from the remote resource.
     */
    public void disconnect();

    /**
     * Connect to the remote resource.
     */
    public void connect() throws IOException;
    
}
