/***********************************************************
 * $Id: Client.java 48 2008-10-29 15:34:24Z wglas $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 15.10.2008
 * 
 ***********************************************************/

package org.clazzes.remoting.client;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.URI;
import java.rmi.ConnectException;
import java.rmi.server.UID;
import java.util.HashMap;
import java.util.Map;


import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.clazzes.remoting.InvalidTargetException;
import org.clazzes.remoting.InvocationHandler;
import org.clazzes.remoting.RemoteInvoker;
import org.clazzes.remoting.cmd.CallBackCmd;
import org.clazzes.remoting.cmd.Cmd;
import org.clazzes.remoting.cmd.CmdVisitorSupport;
import org.clazzes.remoting.cmd.ExchangeClientUIDCmd;
import org.clazzes.remoting.cmd.InvocationCmd;
import org.clazzes.remoting.cmd.InvocationExceptionCmd;
import org.clazzes.remoting.cmd.InvocationResultCmd;
import org.clazzes.remoting.cmd.LoadClassCmd;
import org.clazzes.remoting.cmd.ReturnClassBytesCmd;
import org.clazzes.remoting.helper.InetAddressHelper;
import org.clazzes.remoting.loading.ClassBytes;
import org.clazzes.remoting.loading.SecurityUtility;
import org.clazzes.remoting.marshal.Marshaler;

/**
 * A remote invocation client, which support callbacks over a single TCP/IP.
 * 
 * @author wglas
 */
public class Client implements RemoteInvoker {

    private static final Log log = LogFactory.getLog(Client.class);
    
    private String scheme;
    private InetSocketAddress address;
    private ClientParameters clientParameters;
    private ClassLoader marshalClassLoader;
    private Map <UID,InvocationHandler> callbackHandlers;
    
    // variables used for connection maintainance.
    private ConnectionHandler connectionHandler;
    
    private class ConnectionHandler extends CmdVisitorSupport {

        private final Marshaler marshaler;
        public final UID clientUID;

        public ConnectionHandler(final Marshaler marshaler, final UID clientUID) {
            
            this.marshaler = marshaler;
            this.clientUID = clientUID;
        }
        
        private InvocationCmd invocation;
        private InvocationResultCmd invocationResult;
        private InvocationExceptionCmd invocationException;
        private LoadClassCmd loadClass;
        private ReturnClassBytesCmd returnClassBytes;
        
        private void close() {
            this.marshaler.close();
        }
        
        @Override
        public void visit(CallBackCmd cmd) {
           
            UID uid = cmd.getCallbackUID();
            
            InvocationHandler handler = null;
            
            if (Client.this.callbackHandlers != null)
                synchronized(Client.this.callbackHandlers) {
                    handler = Client.this.callbackHandlers.get(uid);
                }
            
            Cmd response;
            
            if (handler == null) {
                
              response=
                  new InvocationExceptionCmd(
                          new InvalidTargetException("Callback UID ["+uid+"] does not exist or is expired."));
            
            } else {
                
                try {
                    Object res = handler.invoke(Client.this,cmd.getArg(),cmd.getMetadata());
                    
                    response = new InvocationResultCmd(res);
                    
                } catch (Throwable e) {
                    response = new InvocationExceptionCmd(e);
                }
            }
            
            try {
                this.marshaler.writeObject(response);
                
            } catch (IOException e) {
                throw new RuntimeException("I/O exception writing callback response.",e);
            }
        }

        @Override
        public void visit(InvocationExceptionCmd cmd) {
            
            if (this.invocation == null || this.invocationResult != null || this.invocationException != null) {
                super.visit(cmd);
                return;
            }
            
            this.invocationException = cmd;
        }

        @Override
        public void visit(InvocationResultCmd cmd) {
            
            if (this.invocation == null || this.invocationResult != null || this.invocationException != null) {
                super.visit(cmd);
                return;
            }
                
            this.invocationResult = cmd;
        }

        public Object send(InvocationCmd invocation) throws IOException, InvocationTargetException {
            
            if (this.invocation != null || this.loadClass != null)
                throw new IOException("Concurrent remote invocations are not implemented.");
                      
            this.marshaler.writeObject(invocation);
            this.invocation = invocation;
            
            Object obj;
            try {

                while (this.invocationException == null && this.invocationResult == null) {
                    obj = this.marshaler.readObject();
                    Cmd.acceptCmd(obj,this);
                }

            } catch (Throwable e) {
                
                IOException ioe;
                
                if (e instanceof IOException) {
                    ioe = (IOException) e;
                }
                else {
                    ioe = new IOException("Exception caught during read of reponse to remote invocation.");
                    ioe.initCause(e);
                }
                
                throw ioe;
                
            } finally {
                
                this.invocation = null;
            }

            
           if  (this.invocationResult != null) {
                    
               Object ret = this.invocationResult.getResult();
               this.invocationResult = null;
               return ret;
           }
           if  (this.invocationException != null) {
               
               Throwable e = this.invocationException.getException();
               this.invocationException = null;
               throw new InvocationTargetException(e);
           }
           
           throw new IllegalStateException("Illegal state while fetching reponse to remote invocation.");
        }
        
        @Override
        public void visit(ReturnClassBytesCmd cmd) {
            
            if (this.loadClass == null || this.returnClassBytes != null) {
                super.visit(cmd);
                return;
            }
                
            this.returnClassBytes = cmd;
        }

        public ClassBytes send(LoadClassCmd cmd) throws IOException {
            
            if (this.invocation != null || this.loadClass != null)
                throw new IOException("Concurrent remote invocations are not implemented.");
                      
            this.marshaler.writeObject(cmd);
            this.loadClass = cmd;
            

            Object obj;
            try {

                while (this.returnClassBytes == null) {
                    obj = this.marshaler.readObject();
                    Cmd.acceptCmd(obj,this);
                }

            } catch (Throwable e) {
                
                IOException ioe;
                
                if (e instanceof IOException) {
                    ioe = (IOException) e;
                }
                else {
                    ioe = new IOException("Exception caught during read of reponse to class load.");
                    ioe.initCause(e);
                }
                
                throw ioe;
                
           } finally {
                
                this.loadClass = null;
            }
            
            if  (this.returnClassBytes != null) {
                
                ClassBytes ret = this.returnClassBytes.getClassBytes();
                this.returnClassBytes = null;
                return ret;
            }
                
            throw new IllegalStateException("Illegal state while fetching reponse to load class request.");
        }
       
     }
    
    /**
     * Construct a client, which will connect to the given URI.
     * 
     * Please note, that any name resolution required will be performed by
     * {@link Client#connect()}.
     * 
     * The following default parameters will be set:
     * <table>
     *   <tr><td>soTimeout</td><td>60000</td></tr>
     *   <tr><td>tcpNoDelay</td><td>true</td></tr>
     * </table>
     * 
     * @param uri An URI with a supported scheme. Currently, only the <code>tcp</code>
     *           and <code>ssl</code> schemes are supported.
     * @throws IOException If the URI has an unsupported scheme.
     */
    public Client(URI uri) throws IOException {
        
        this.scheme = uri.getScheme();
        
        if ((!"tcp".equals(this.scheme)) && (!"ssl".equals(this.scheme)))
            throw new IOException("Scheme ["+this.scheme+"] is not supported.");
        
        this.address = InetSocketAddress.createUnresolved(uri.getHost(),uri.getPort());
    }
    
    /**
     * @return The client configuration object. If not set via {@link #setClientParameters(ClientParameters)},
     *           a new instance of {@link ClientParameters} or {@link SSLClientParameters} is returned
     *           depending on the scheme of the URI passed to the constructor.
     */
    public ClientParameters getClientParameters() {
        
        if (this.clientParameters == null) {
            
        if ("ssl".equals(this.scheme)) 
            this.clientParameters = new SSLClientParameters();
        else
            this.clientParameters = new ClientParameters();
        }
        
        return this.clientParameters;
    }

    /**
     * @param clientParameters The client parameters to set.
     */
    public void setClientParameters(ClientParameters clientParameters) {
        this.clientParameters = clientParameters;
    }

    /**
     * @return The class loader used for resolving deserialized objects. If not
     *         set via {@link #setMarshalClassLoader(ClassLoader)}, the context class
     *         loader of the current thread is used.
     */
    public ClassLoader getMarshalClassLoader() {
        
        if (this.marshalClassLoader == null)
            this.marshalClassLoader = SecurityUtility.getContextClassLoader(Thread.currentThread());
        
        return this.marshalClassLoader;
    }

    /**
     * @param classLoader The custom class loader used for resolving deserialized objects.
     */
    public void setMarshalClassLoader(ClassLoader classLoader) {
        this.marshalClassLoader = classLoader;
    }
    
    private static class GetClientUIDVisitor extends CmdVisitorSupport {
        
        UID clientUID;
        
        @Override
        public void visit(ExchangeClientUIDCmd cmd) {
            this.clientUID = cmd.getClientUID();
        }
    }
    
    /* (non-Javadoc)
     * @see org.clazzes.remoting.RemoteInvoker#connect()
     */
    public synchronized void connect() throws IOException {
        
        if (this.connectionHandler != null)
            throw new ConnectException("Try to connect an already conencted client.");
        
        Marshaler m = this.getClientParameters().getMarshalerFactory().newMarshaler();

        try {
            Socket socket =
                this.getClientParameters().getSocketFactory().createSocket();
            
            // Setting an soTimeout during connection establishment helps to bail
            // out firewall problems an the like.
            //
            // Specifically, a connection establishment with a SYN packet that is never
            // answered by a FIN or ACK package typically blocks for minutes in an operating
            // system dependent manner.
            //
            // Additionally, if a server is not able to communicate the client UID in a timely
            // fashion, the server may be overloaded or deadlocked. 
            //
            this.getClientParameters().tuneSocket(socket);
            
            InetSocketAddress addr =
                InetAddressHelper.resolveSocketAddress(this.address);
            
            if(log.isDebugEnabled())
                log.debug("Connecting to ["+addr+"]...");
            
            SecurityUtility.connect(socket,addr,this.getClientParameters().getSoTimeout());
        
            m.setup(socket.getOutputStream(),socket.getInputStream(),
                    this.getMarshalClassLoader());
     
            // Fetch the client UID from the server during an initial handshake.
            Object obj = m.readObject();
            
            GetClientUIDVisitor v = new GetClientUIDVisitor();
            
            Cmd.acceptCmd(obj,v);
            
            // The ConnectionHandler is not able to distinguish between a long lasting
            // remote call and a real socket timeout, so we set the socket to blocking
            // operation from here on.
            socket.setSoTimeout(0);
            
            this.connectionHandler = new ConnectionHandler(m,v.clientUID);
            
            log.info("Client ["+v.clientUID+"] has been successfully connected to ["+addr+"].");
            
            m = null;
            
        } catch (ClassNotFoundException e) {
            
            log.error("Class resolution error in startup handshake.",e);
            throw new IOException("Class resolution error in startup handshake.");
            
        } finally {
            
            if (m != null) {
                m.close();
                this.connectionHandler = null;
            }
        }
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.RemoteInvoker#deregisterCallback(java.rmi.server.UID)
     */
    public void deregisterCallback(UID handlerUID) {
        
        InvocationHandler handler = null;
        
        if (this.callbackHandlers != null) {
        
            synchronized(this.callbackHandlers) {
                handler = this.callbackHandlers.remove(handlerUID);
            }
        }
        
        if (handler == null)
            log.warn("Trying to deregister callback for non-existent UID ["+handlerUID+"]");
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.RemoteInvoker#disconnect()
     */
    public synchronized void disconnect() {
        
        if (this.connectionHandler != null) {
        
            UID clientUID = this.connectionHandler.clientUID;
            
            this.connectionHandler.close();
            this.connectionHandler = null;
            
            log.info("Client ["+clientUID+"] has been successfully disonnected from ["+this.address+"].");

            this.connectionHandler = null;
        }
        else {
            log.warn("Try to diconnect a client from ["+this.address+"] which is already disconnected.");
        }
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.RemoteInvoker#invoke(java.lang.String, java.lang.Object, java.util.Map)
     */
    public Object invoke(String subSystem, Object arg,
            Map<String, Object> metadata) throws InvocationTargetException,
            IOException {
       
        ConnectionHandler handler;
          
        // avoid lock over the whole span of a remote call, since
        // then long-lasting call might prevent us from shutting down a client
        // server system.
        synchronized(this) {
        
            if (this.connectionHandler == null)
                throw new IOException("Invocation on an unconnected client.");
            
            handler = this.connectionHandler;
        }
        
        InvocationCmd cmd = new InvocationCmd(subSystem,arg,metadata);

        return handler.send(cmd);
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.RemoteInvoker#loadClass(java.lang.String)
     */
    public ClassBytes loadClass(String className) throws IOException {
        
        ConnectionHandler handler;
        
        // avoid lock over the whole span of a remote call, since
        // then long-lasting call might prevent us from shutting down a client
        // server system.
        synchronized(this) {
        
            if (this.connectionHandler == null)
                throw new IOException("Try to load a class on an unconnected client.");
            
            handler = this.connectionHandler;
        }
        
        LoadClassCmd cmd = new LoadClassCmd(className);

        return handler.send(cmd);
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.RemoteInvoker#isConnected()
     */
    public synchronized boolean isConnected() {
        
        return this.connectionHandler != null;
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.RemoteInvoker#registerCallback(org.clazzes.remoting.InvocationHandler)
     */
    public synchronized UID registerCallback(InvocationHandler handler) throws IOException {
        
        if (this.callbackHandlers == null)
            this.callbackHandlers = new HashMap<UID, InvocationHandler>();
            
        UID uid = new UID();

        synchronized(this.callbackHandlers) {
               
            this.callbackHandlers.put(uid,handler);
        }
        
        return uid; 
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.InvocationContext#getClientUID()
     */
    public synchronized UID getClientUID() {
        
        return this.connectionHandler == null ? null : this.connectionHandler.clientUID;
    }
}
