/***********************************************************
 * $Id: ClientPool.java 48 2008-10-29 15:34:24Z wglas $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 15.10.2008
 * 
 ***********************************************************/

package org.clazzes.remoting.client;

import java.io.IOException;
import java.net.URI;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Queue;
import java.util.concurrent.TimeoutException;

import javax.security.auth.DestroyFailedException;
import javax.security.auth.Destroyable;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.clazzes.remoting.loading.ClassByteClassLoader;
import org.clazzes.remoting.loading.SecurityUtility;

/**
 * A thread-safe pool of clients.
 * 
 * @author wglas
 */
public class ClientPool implements Destroyable, ClientManager {

	private static final Log log = LogFactory.getLog(ClientPool.class);
		
	private static class ClientEntry {

		public final int id;
		public final Client client;
		public long lastChanged;
		
		public ClientEntry(int id, Client client, long lastChanged) {
			super();
			this.id = id;
			this.client = client;
			this.lastChanged = lastChanged;
		}
	}
	
	private Map<Client,ClientEntry> activeClients;
	private Queue<ClientEntry> idleClients;
	
	private int lastClientId;
	
	private int minPoolSize;
	private int maxPoolSize;
	private long idleTimeout;
	private long fetchTimeout;
	private boolean loadRemoteClasses;

	private ClientParameters clientParameters;
	
	private URI connectURI;

	private Runnable gcRunnable;
	private boolean pendingShutdown;
    private ClassLoader marshalClassLoader;
	
	private ClassLoader getMarshalClassLoader() throws IOException {
	    
	    if (this.marshalClassLoader == null) {
	        
	        ClassLoader ctxtLoader = SecurityUtility.getContextClassLoader(Thread.currentThread());
	        
	        if (this.loadRemoteClasses) {
	        
	            ClassByteClassLoader cbLoader = new ClassByteClassLoader(ctxtLoader);
	            
	            Client clClient = new Client(this.connectURI);
	            
	            clClient.setMarshalClassLoader(ctxtLoader);
	            clClient.setClientParameters(this.getClientParameters());
	            
	            cbLoader.setClientInvoker(clClient);
	            
	            this.marshalClassLoader = cbLoader;
	        }
	        else {
	            this.marshalClassLoader = ctxtLoader;
	        }
	    }
	    
	    return this.marshalClassLoader;
	}
	
	// not synchronized, it is assumed, that this method is ever called from
	// within another synchronized context.
	private ClientEntry createClient() throws IOException {
		
	    ClassLoader loader = this.getMarshalClassLoader();
	    
		Client client = new Client(this.connectURI);
		
		client.setClientParameters(this.getClientParameters());
		client.setMarshalClassLoader(loader);
		
		int id = ++this.lastClientId;
		
		if (log.isDebugEnabled())
			log.debug("Client id ["+id+"] created.");
			
		return new ClientEntry(id,client,System.currentTimeMillis());
	}
	
	private void disposeClient(ClientEntry clientEntry)  {
		
		Client client = clientEntry.client;
				
		if (log.isDebugEnabled())
			log.debug("Disposing client id ["+clientEntry.id+"].");

		if (client.isConnected()) {
		
			try {
				client.disconnect();
			} catch (Throwable e) {
				log.warn("Caught Exception disconnecting client id ["+clientEntry.id+"].",e);
			}
		}
	}
	
	private synchronized long gc() {
		
		if (this.pendingShutdown) return -1;
		
		int ndisposable = this.activeClients.size() + this.idleClients.size() - this.minPoolSize;
		
		if (log.isDebugEnabled())
			log.debug("gc: ["+ndisposable+"] clients may be considered.");

		long waitTime = this.idleTimeout;
		
		if (ndisposable > 0) {
		
			long now = System.currentTimeMillis();
		
			while (this.idleClients.size() > 0 && ndisposable > 0) {
			
				ClientEntry clientEntry = this.idleClients.peek();
				
				long idleDuration = now - clientEntry.lastChanged;
				
				if (log.isDebugEnabled())
					log.debug("gc: client id ["+clientEntry.id+"] is idle since ["+idleDuration+"] ms.");

				if (idleDuration < this.idleTimeout) {
					
					long delta = this.idleTimeout - idleDuration;
					waitTime = delta < 100L ? 100L : delta;
					break;
				}
			
				this.disposeClient(clientEntry);
				this.idleClients.poll();
					
				--ndisposable;
			}
		}
		
		return waitTime;
	}

	private class GCRunnable implements Runnable  {

		/* (non-Javadoc)
		 * @see java.lang.Runnable#run()
		 */
		public void run() {

			try {
				long waitTime;
				
				log.info("GC Thread has been started.");

				while ((waitTime=ClientPool.this.gc()) >= 0)  {
					try {

						synchronized (this) {
						
							if (log.isDebugEnabled())
								log.debug("gc: waiting for ["+waitTime+"] ms.");
							
							this.wait(waitTime);
						}
					} catch (InterruptedException e) {
						log.warn("Caught interrupt in garbage collector.",e);
					}
				}
				
				log.info("GC Thread has been stopped.");
				
			} catch (Throwable e) {
				ClientPool.log.error("Caught exception in garbage collector thread.",e);
			}
		}
	}
	
	/**
	 * Construct a client pool with the default parameters
	 * <table>
     *   <tr><td>minPoolSize</td><td>4</td></tr>
     *   <tr><td>maxPoolSize</td><td>8</td></tr>
     *   <tr><td>idleTimeout</td><td>60000</td></tr>
     *   <tr><td>fetchTimeout</td><td>30000</td></tr>
     *   <tr><td>soTimeout</td><td>60000</td></tr>
     *   <tr><td>tcpNoDelay</td><td>true</td></tr>
	 * </table>
	 */
	public ClientPool() {
		
		this.minPoolSize = 4;
		this.maxPoolSize = 8;
		this.idleTimeout = 60000L;
		this.fetchTimeout = 3000L;
		this.pendingShutdown = false;
		this.lastClientId = 0;
	}
	
    /**
     * @return The client configuration object. If not set via {@link #setClientParameters(ClientParameters)},
     *           a new instance of the base class {@link ClientParameters} is returned.
     */
    public ClientParameters getClientParameters() {
        
        if (this.clientParameters == null)
            this.clientParameters = new ClientParameters();
        
        return this.clientParameters;
    }

    /**
     * @param clientParameters The client parameters to set.
     */
    public void setClientParameters(ClientParameters clientParameters) {
        this.clientParameters = clientParameters;
    }

    /* (non-Javadoc)
     * @see javax.security.auth.Destroyable#isDestroyed()
     */
    public synchronized boolean isDestroyed() {
       
        return this.idleClients == null;
    }

    /* (non-Javadoc)
     * @see javax.security.auth.Destroyable#destroy()
     */
    /* (non-Javadoc)
     * @see org.clazzes.remoting.client.ClientManager#destroy()
     */
    public synchronized void destroy() throws DestroyFailedException {
 		
		if (this.idleClients == null) return;
		
		this.pendingShutdown = true;
		// notify the gc thread, so it will stop.
		synchronized (this.gcRunnable) {
			this.gcRunnable.notify();
		}
		this.notify();
		
		while (this.activeClients.size() > 0) {
			
			log.info("There are ["+this.activeClients.size()+"] clients busy, waiting for all clients to come down...");
			
			try {
                this.wait();
            } catch (InterruptedException e) {
                throw new DestroyFailedException("");
            }
		}
		
		log.info("Shutting down ["+this.idleClients.size()+"] remaining clients.");
		
		for (ClientEntry clientEntry : this.idleClients) {
			
			this.disposeClient(clientEntry);
		}
		
		if (this.marshalClassLoader instanceof ClassByteClassLoader) {
		 
		    ((ClassByteClassLoader)this.marshalClassLoader).destroy();
		}
		
		this.marshalClassLoader = null;
		this.activeClients = null;
		this.idleClients = null;
		this.gcRunnable = null;
	}

	/* (non-Javadoc)
     * @see org.clazzes.remoting.client.ClientManager#start()
     */
	public synchronized void start() throws IOException {
		
		this.idleClients = new LinkedList<ClientEntry>();
		this.activeClients = new HashMap<Client, ClientEntry>(this.maxPoolSize);
		 	
		for (int i = 0; i<this.minPoolSize; ++i) {
			
			ClientEntry clientEntry = this.createClient();
			
			this.idleClients.add(clientEntry);
		}
		
		this.gcRunnable = new GCRunnable();
		Thread thread = new Thread(this.gcRunnable);
		thread.start();
		
		this.pendingShutdown = false;
	}

	/* (non-Javadoc)
     * @see org.clazzes.remoting.client.ClientManager#activateClient()
     */
	public synchronized Client activateClient() throws TimeoutException {
		
		if (this.pendingShutdown) throw new TimeoutException("Try to fetch a Client during shutdown.");
		
		if (this.idleClients.size() <= 0 && this.activeClients.size() >= this.maxPoolSize) {
			
			try {
				this.wait(this.fetchTimeout);
				if (this.pendingShutdown) throw new TimeoutException("Try to fetch a Client during shutdown.");
				
			} catch (InterruptedException e) {
				log.warn("Caught interrupt waiting for a client to become available.");
			}
		}
			
		if (this.idleClients.size() <= 0 && this.activeClients.size() < this.maxPoolSize) {
				
			try {
				this.idleClients.add(this.createClient());
			} catch (Exception e) {
				log.fatal("Error creating a new Client.",e);
				throw new TimeoutException("Error creating a new Client.");
			}
		}
		
		if (this.idleClients.size() > 0) {
			
			ClientEntry entry = this.idleClients.poll();
			
			entry.lastChanged = System.currentTimeMillis();
			
			if (!entry.client.isConnected()) {
				try {
					entry.client.connect();
					
				} catch (Throwable e) {
					this.idleClients.add(entry);
					
					log.error("Error connecting an idle Client.",e);
					throw new TimeoutException("Error connecting a not yet connected Client.");
				}
			}
			
			this.activeClients.put(entry.client,entry);
			
			if (log.isDebugEnabled())
				log.debug("Activated client id ["+entry.id+"].");
			
			return entry.client;
		}
		
		throw new TimeoutException("No more clients available, try to increase the maximal pool size.");
	}
	
	/* (non-Javadoc)
     * @see org.clazzes.remoting.client.ClientManager#passivateClient(org.clazzes.remoting.client.Client)
     */
	public synchronized void passivateClient(Client client) {
		
		ClientEntry entry = this.activeClients.remove(client);
		
		if (entry == null)
			throw new NullPointerException("The put back client is not managed by this pool.");
		
		
		if (entry.client.isConnected()) {
			
			entry.lastChanged = System.currentTimeMillis();
			this.idleClients.add(entry);
			
			if (log.isDebugEnabled())
				log.debug("Passivated client id ["+entry.id+"].");
		} else {
			
			this.disposeClient(entry);
			
			try {
				while (this.activeClients.size() + this.idleClients.size() < this.minPoolSize) {
					this.idleClients.add(this.createClient());
				}
			} catch (Exception e) {
				log.error("Caught Exception creating a new idle client.",e);
			}
		}
		
		this.notify();
	}
	
    /**
	 * @return the minimal size of the client pool.
	 *         The default value is <code>4</code>.
	 */
	public int getMinPoolSize() {
		return this.minPoolSize;
	}

	/**
	 * @param minPoolSize the minimal size of the client pool to set.
	 */
	public void setMinPoolSize(int minPoolSize) {
		this.minPoolSize = minPoolSize;
	}

	/**
	 * @return the maximal size of the client pool.
	 *         The default value is <code>8</code>.
	 */
	public int getMaxPoolSize() {
		return this.maxPoolSize;
	}

	/**
	 * @param maxPoolSize the maximal size of the client pool to set.
	 */
	public void setMaxPoolSize(int maxPoolSize) {
		this.maxPoolSize = maxPoolSize;
	}

	/**
	 * @return The timeout in milliseconds after which idle clients are disposed.
	 *         The default value is <code>60000</code>.
	 */
	public long getIdleTimeout() {
		return this.idleTimeout;
	}

	/**
	 * @param idleTimeout the idleTimeout to set
	 */
	public void setIdleTimeout(long idleTimeout) {
		this.idleTimeout = idleTimeout;
	}

	/**
	 * @return The URI to which the generated clients will connect.
	 */
	public URI getConnectURI() {
        return this.connectURI;
    }

    /**
     * @param connectURI The connect URI to set.
     */
    public void setConnectURI(URI connectURI) {
        this.connectURI = connectURI;
    }

    /**
     * @return Whether to load classes from the remote server on demand.
     */
    public boolean isLoadRemoteClasses() {
        return this.loadRemoteClasses;
    }

    /**
     * @param loadRemoteClasses Whether to load classes from the remote server on demand.
     */
    public void setLoadRemoteClasses(boolean loadRemoteClasses) {
        this.loadRemoteClasses = loadRemoteClasses;
    }

    /**
	 * @return the maximal time in milliseconds to wait until a client becomes available.
	 *         The default value is <code>3000</code>.
	 */
	public long getFetchTimeout() {
		return this.fetchTimeout;
	}


	/**
	 * @param fetchTimeout the fetchTimeout to set
	 */
	public void setFetchTimeout(long fetchTimeout) {
		this.fetchTimeout = fetchTimeout;
	}
	
}
