/***********************************************************
 * $Id: JAVAMarshaler.java 48 2008-10-29 15:34:24Z wglas $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 15.10.2008
 * 
 ***********************************************************/

package org.clazzes.remoting.marshal;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.EOFException;
import java.io.FilterInputStream;
import java.io.FilterOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.util.zip.GZIPOutputStream;

import javax.crypto.CipherInputStream;
import javax.crypto.CipherOutputStream;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.clazzes.remoting.loading.ObjectInputStreamWithClassLoader;

/**
 * The default implementation of the marshaler, which uses an
 * {@link ObjectInputStreamWithClassLoader} and an {@link ObjectOutputStream}
 * for deserialization and serialization.
 * 
 * @author wglas
 */
public class JAVAMarshaler implements Marshaler {
    
    private static final Log log = LogFactory.getLog(JAVAMarshaler.class);
    
    protected ObjectInputStream ois;
    protected ObjectOutputStream oos;
    private int outputBufferSize;
    private int inputBufferSize;
    
    /**
     * Default constructor, which set the output and input buffer size to 0.
     */
    public JAVAMarshaler() { 
    }
    
    /**
     * Construct a JAVAMarshaler, with given input and output buffer sizes. 
     * @param outputBufferSize The size of the output buffer to use.
     * @param inputBufferSize The size of the input buffer to use.
     */
    public JAVAMarshaler(int outputBufferSize, int inputBufferSize) {
        super();
        this.outputBufferSize = outputBufferSize;
        this.inputBufferSize = inputBufferSize;
    }

    /**
     * Prepare the transport output stream for use with ObjectOutputStream.
     * This may involve wrapping the stream with a {@link BufferedInputStream} or
     * some other sort of {@link FilterInputStream} like {@link CipherInputStream} or
     * {@link GZIPOutputStream}.
     * 
     * The default implementation wraps the output stream with {@link BufferedInputStream},
     * if the output buffer size is set to a positive value.
     * 
     * @param os The raw OutputStream as passed to {@link #setup(OutputStream, InputStream, ClassLoader)}.
     * @return A wrapped output stream.
     * @throws IOException Upon errors setting up the stream.
     */
    protected OutputStream prepareOutputStream(OutputStream os) throws IOException {
        if (this.outputBufferSize <= 0)
            return os;
        else
            return new BufferedOutputStream(os,this.outputBufferSize);
    }
    
    /**
     * @param os The output stream returned by {@link #prepareOutputStream(OutputStream)}.
     * @return An object output stream for writing objects.
     * @throws IOException Upon errors.
     */
    protected ObjectOutputStream setupObjectOutputStream(OutputStream os) throws IOException {
        
        ObjectOutputStream ret = new ObjectOutputStream(os);
        // flush the object output stream here in order to transport the stream
        // header of the network. Otherwise, the InputStream waits infinitely on the stream when
        // using buffered output streams.
        ret.flush();
        return ret;
    }
    
    /**
     * Prepare the transport input stream for use with ObjectInputStream.
     * This may involve wrapping the stream with a {@link BufferedOutputStream} or
     * some other sort of {@link FilterOutputStream} like {@link CipherOutputStream} or
     * {@link GZIPOutputStream}.
     * 
     * The default implementation wraps the input stream with {@link BufferedOutputStream},
     * if the input buffer size is set to a positive value.
     * 
     * 
     * @param is The raw InputStream as passed to {@link #setup(OutputStream, InputStream, ClassLoader)}.
     * @return A wrapped InputStream.
     */
    protected InputStream prepareInputStream(InputStream is) throws IOException {
        if (this.inputBufferSize <= 0)
            return is;
        else
            return new BufferedInputStream(is,this.inputBufferSize);
    }
    
    /**
     * @param is The input stream as return by {@link #prepareInputStream(InputStream)}.
     * @param inputClassLoader The class loader to use for deserialization.
     * @return An ObjectInputStream for the marshaler.
     * @throws IOException Upon errors setting up the ObjectInputStream.
     */
    protected ObjectInputStream setupObjectInputStream(InputStream is, ClassLoader inputClassLoader)
    throws IOException {
        
        return new ObjectInputStreamWithClassLoader(is,inputClassLoader);
    }
        
    /* (non-Javadoc)
     * @see org.clazzes.remoting.marshal.Marshaler#setup(java.io.OutputStream, java.io.InputStream, java.lang.ClassLoader)
     */
    public void setup(OutputStream os, InputStream is, ClassLoader inputClassLoader)
    throws IOException {
        
        try {
            this.oos = this.setupObjectOutputStream(this.prepareOutputStream(os));
            this.ois = this.setupObjectInputStream(this.prepareInputStream(is),inputClassLoader);
            
        } catch (IOException e) {
            
            is.close();
            
            if (this.oos != null) {
                try {
                    this.oos.close();
                } catch (Exception e2) {
                    log.warn("Error while closing output stream upon input stream setup problem",e2);
                }
                this.oos = null;
            } else
                os.close();
            
            throw e;
        }
    }
    
    /* (non-Javadoc)
     * @see org.clazzes.remoting.marshal.Marshaller#readObject()
     */
    public Object readObject() throws ClassNotFoundException, IOException {
        
        if (this.ois == null)
            throw new EOFException("Try to read from a closed marshaler.");

        return this.ois.readObject();
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.marshal.Marshaller#reset()
     */
    public void reset() throws IOException {
        
        if (this.oos == null)
            throw new EOFException("Try to reset a closed marshaler.");

        this.oos.reset();
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.marshal.Marshaller#writeObject(java.lang.Object)
     */
    public void writeObject(Object obj) throws IOException {
        
        if (this.oos == null)
            throw new EOFException("Try to write to a closed marshaler.");
            
        this.oos.writeObject(obj);
        this.oos.reset();
        this.oos.flush();
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.marshal.Marshaller#close()
     */
    public void close() {
        
        try {
            if (this.ois != null)
                this.ois.close();
        } catch (Throwable e) {
            log.warn("Error closing object input stream",e);
        }
        this.ois = null;
        try {
            if (this.oos != null)
            this.oos.close();
        } catch (Throwable e) {
            log.warn("Error closing object output stream",e);
        }
        this.oos = null;
    }

    /**
     * @return The buffer size to use for the {@link BufferedOutputStream} used.
     *         If set to 0, no output buffering is performed.
     */
    public int getOutputBufferSize() {
        return this.outputBufferSize;
    }

    /**
     * @param outputBufferSize The output buffer size to set.
     */
    public void setOutputBufferSize(int outputBufferSize) {
        this.outputBufferSize = outputBufferSize;
    }

    /**
     * @return The buffer size to use for the {@link BufferedInputStream} used.
     *         If set to 0, no input buffering is performed.
     */
    public int getInputBufferSize() {
        return this.inputBufferSize;
    }

    /**
     * @param inputBufferSize The input buffer size to set.
     */
    public void setInputBufferSize(int inputBufferSize) {
        this.inputBufferSize = inputBufferSize;
    }

}
