/***********************************************************
 * $Id: SSLServerParameters.java 48 2008-10-29 15:34:24Z wglas $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 24.10.2008
 * 
 ***********************************************************/

package org.clazzes.remoting.server;

import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.ArrayList;
import java.util.List;

import javax.net.ServerSocketFactory;
import javax.net.ssl.HandshakeCompletedListener;
import javax.net.ssl.SSLServerSocket;
import javax.net.ssl.SSLServerSocketFactory;
import javax.net.ssl.SSLSocket;

/**
 * An SSL-specific implementation of the server parameters object.
 * 
 * @author wglas
 */
public class SSLServerParameters extends ServerParameters {

    private List<HandshakeCompletedListener> handshakeCompletedListeners;
    private boolean needClientAuth;
    private boolean wantClientAuth;
    private boolean enableSessionCreation;
    
    /**
     * Construct an SSL server parameters object with the following default parameters:
     * <table>
     *   <tr><td>needClientAuth</td><td>true</td></tr>
     *   <tr><td>wantClientAuth</td><td>false</td></tr>
     *   <tr><td>enableSessionCreation</td><td>true</td></tr>
     *   <tr><td>tcpNoDelay</td><td>true</td></tr>
     *   <tr><td>backlog</td><td>4</td></tr>
     * </table>
     */
    public SSLServerParameters() {
       
        super();
        
        this.needClientAuth = true;
        this.wantClientAuth = false;
        this.enableSessionCreation = true;
    }
    
    /**
     * Return the socket factory used. If not set via {{@link #setServerSocketFactory(ServerSocketFactory)},
     * {@link SSLServerSocketFactory#getDefault()} is returned.
     */
    @Override
    public ServerSocketFactory getServerSocketFactory() {
       
        if (super.serverSocketFactory == null)
            super.serverSocketFactory = SSLServerSocketFactory.getDefault();
        
        return super.serverSocketFactory;
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.client.ClientParameters#tuneSocket(java.net.Socket)
     */
    @Override
    public void tuneSocket(Socket socket) throws IOException {
       
        if (socket instanceof SSLSocket) {
        
            SSLSocket sslSocket = (SSLSocket) socket;
            
            if (this.handshakeCompletedListeners != null)
                for (HandshakeCompletedListener listener : this.handshakeCompletedListeners)
                    sslSocket.addHandshakeCompletedListener(listener);
            
        }
        super.tuneSocket(socket);
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.server.ServerParameters#tuneServerSocket(java.net.ServerSocket)
     */
    @Override
    public void tuneServerSocket(ServerSocket serverSocket) {
     
        if (serverSocket instanceof SSLServerSocket) {
            SSLServerSocket sslServerSocket = (SSLServerSocket) serverSocket;
            
            if (this.needClientAuth)
                sslServerSocket.setNeedClientAuth(this.needClientAuth);
            else
                sslServerSocket.setWantClientAuth(this.wantClientAuth);
            
            sslServerSocket.setEnableSessionCreation(this.enableSessionCreation);
        }
        
        super.tuneServerSocket(serverSocket);
    }

    /**
     * @return The list of SSL handshake completion listeners to register in
     *     {@link #tuneSocket(Socket)}.
     */
    public List<HandshakeCompletedListener> getHandshakeCompletedListeners() {
        return this.handshakeCompletedListeners;
    }

    /**
     * @param handshakeCompletedListeners The list of SSL handshake completion
     *   listeners to register in {@link #tuneSocket(Socket)} to set.
     */
    public void setHandshakeCompletedListeners(
            List<HandshakeCompletedListener> handshakeCompletedListeners) {
        this.handshakeCompletedListeners = handshakeCompletedListeners;
    }
    
    /**
     * @param handshakeCompletedListener An SSL handshake completion
     *   listener to be registered by {@link #tuneSocket(Socket)} to add.
     */
    public void addHandshakeCompletedListeners(
            HandshakeCompletedListener handshakeCompletedListener) {
        
        if (this.handshakeCompletedListeners == null)
            this.handshakeCompletedListeners = new ArrayList<HandshakeCompletedListener>();
        
        this.handshakeCompletedListeners.add(handshakeCompletedListener);
    }

    /**
     * @return Whether the SSL socket requests a obligatory client authentication.
     * 
     * @see SSLServerSocket#getNeedClientAuth()
     */
    public boolean isNeedClientAuth() {
        return this.needClientAuth;
    }

    /**
     * @param needClientAuth The need client authentication flag of the SSL socket to set.
     * 
     * @see SSLServerSocket#setNeedClientAuth(boolean)
     */
    public void setNeedClientAuth(boolean needClientAuth) {
        this.needClientAuth = needClientAuth;
    }

    /**
     * @return Whether the SSL socket requests an optional client authentication.
     *         If {@link #isNeedClientAuth()} returns <code>true</code>, this flag
     *         is ignored.
     * 
     * @see SSLServerSocket#getWantClientAuth()
     */
    public boolean isWantClientAuth() {
        return this.wantClientAuth;
    }

    /**
     * @param wantClientAuth The request client authentication flag of the SSL socket to set.
     *         If {@link #isNeedClientAuth()} returns <code>true</code>, this flag
     *         is ignored.
     * 
     * @see SSLServerSocket#setWantClientAuth(boolean)
     */
    public void setWantClientAuth(boolean wantClientAuth) {
        this.wantClientAuth = wantClientAuth;
    }

    /**
     * @return Whether to enable session creation on the generated SSL socket.
     * 
     * @see SSLServerSocket#getEnableSessionCreation()
     */
    public boolean isEnableSessionCreation() {
        return this.enableSessionCreation;
    }

    /**
     * @param enableSessionCreation The session enable flag to set to the SSL socket.
     * 
     * @see SSLServerSocket#setEnableSessionCreation(boolean)
     */
    public void setEnableSessionCreation(boolean enableSessionCreation) {
        this.enableSessionCreation = enableSessionCreation;
    }
}
